# let's get some useful pieces from other files
from tools import *                # provides useful DNA/RNA/Protein conversions
from fasta import parse_fasta

# let's set some configuration parameters
MINSIZE = 150
UPSTREAM_AMT = 45

FILE = 'X06403.1.fasta'
CIRCULAR = True


def find_ORF_primary(dna):
    """Return list of ORFs on primary strand, using NCBI (start,stop) values."""
    n = len(dna)
    if CIRCULAR:
        dna = 2*dna
        
    orfs = []
    for frame in (0,1,2):   # restart process with each of the three frames
        k = frame
        active = False
        while k+2 < len(dna):
            codon = dna[k:k+3]
            if codon == START_CODON and k < n and not active:
                j = k             # record starting location
                active = True
            elif active and codon in STOP_CODONS:
                active = False
                if k - j + 1 >= MINSIZE:
                    orfs.append( (j+1,k+3) )     # NCBI style numbering
            k += 3
    return orfs
                    
def find_ORF_secondary(dna):
    """Return list of ORFs on primary strand, using NCBI (start,stop) values."""
    n = len(dna)
    orfs = []
    temp = find_ORF_primary(reverse_complement(dna))
    for (a,b) in temp:
        orfs.append( (n+1-a, n+1-b) )
    return orfs
                    
def find_ORFs(dna):
    """Return a list of all valid ORFs for the dna."""
    orfs = find_ORF_primary(dna) + find_ORF_secondary(dna)
    return orfs

def orf_length(orf):
    """Return the ORF length measured in nucleotides"""
    return 1+abs(orf[0]-orf[1])

def orf_is_primary(orf):
    """Return True if orf is on the primary strand,
       or False if on the complementary strand."""
    return orf[0] < orf[1]

def orf_ref_frame(dna, orf):
    """Return the reference frame containing the orf
       (either 1, 2, or 3, as per NCBI style)."""
    if orf_is_primary(orf):
        return 1 + ((orf[0]+2) % 3)
    else:
        return 1 + ( (len(dna)-orf[0]) % 3)

def upstream(dna, orf, k):
    """Return the k adjacent nucleotides on the 5' end of the given orf.

    The result is ordered starting from most upstream to least upstream.
    """
    if orf_is_primary(orf):
        start = max(0,orf[0]-k-1)
        return dna[start:orf[0]-1]
    else:
        raw = dna[orf[0]:orf[0]+k]
        return reverse_complement(raw)

def downstream(dna, orf, k):
    """Return the k adjacent nucleotides on the 3' end of the given orf.

    The result is ordered starting from least downstream to most downstream.
    """
    if orf_is_primary(orf):
        return dna[orf[1]:orf[1]+k]
    else:
        raw = dna[orf[1]-k-1:orf[1]-1]
        return reverse_complement(raw)

def format_orf(dna, orf):
    """Return a string description of the orf, akin to NCBI ORF Finder data."""
    strand = '+' if orf_is_primary(orf) else '-'
    frame = orf_ref_frame(dna, orf)
    length = orf_length(orf)
    return '  %s      %d   %5d %5d  %5d|%d' % (
        strand, frame, orf[0], orf[1], length, length//3 - 1)


#------------------------------------------------------------
# Here is where the top-level work begins...
#------------------------------------------------------------
dna = parse_fasta(FILE)
orfs = [orf for orf in find_ORFs(dna) if orf_length(orf) >= MINSIZE]
orfs.sort(key=orf_length, reverse=True)
print('%d ORFs found' %  len(orfs))
print('strand frame start  stop  length(nt|aa)  upstream')
for orf in orfs:
    print('%-40s %s' % (format_orf(dna,orf), upstream(dna,orf,UPSTREAM_AMT)))
